from time import time

from funki import lambdas
from funki.terminal import Terminal
from funki import report
from funki import text_colours as tc

default_equality_operator = lambda y, z: y==z

class Funk:
  def _default_run(self, state, x):
    print()
    raise NotImplementedError(
      f"please specify run for funk: '{tc.info(self.name)}'"
    )

  def __init__(
    self,
    name:str,
    io_value_pairs=[],
    output_properties=[],
    description='',
    setup=lambdas.default_setup,
    run=None,
    tear_down=lambdas.nop_x,
    custom_tests=None,
    special_return_equality_operator=None,
    enabled=True
  ):
    self.name = name
    self._state = {'term': Terminal()}
    self.description = description
    self._run = self._default_run if run is None else run
    self._tear_down = tear_down

    setup_result = setup()
    if setup_result is None:
      setup_result = {}

    for key in setup_result.keys():
      self._state[key] = setup_result[key]

    if all([
      not io_value_pairs,
      not output_properties,
      not custom_tests
    ]):
      raise ValueError(
        'Must specify io_value pairs, output properties, or custom_tests.'
      )

    self.io_value_pairs = io_value_pairs
    self._output_properties = output_properties
    self.custom_tests = custom_tests

    if special_return_equality_operator is not None:
      self.equality_operator = special_return_equality_operator
    else:
      self.equality_operator = default_equality_operator
  
    self.enabled = enabled
    
  run = lambda self, x=None: self._run(self._state, x)

  def _test(self, max_filename_length):
    """
    Input x provided to function f returns obervation z
    x --> f(x) --> z

    Input x provided to function f is expected to return y
    x --> f(x) --> y

    Test passes is expected return value y is equal to observation z.

    x includes:
      x_args: arguments passed to the function as single value or dict of values
      x_stdi: stdi stream provided to the function while running
    
    y includes:
      y_return: the expected return value from calling f(x)
      y_stdo: the expected stdo data emitted by f(x) during it's execution.
    
    z includes:
      z_return: the observed return value from calling f(x)
      z_stdo: the observed stdo data emitted by f(x) during it's execution.
    """

    test_start_time = time()

    # For each z:

    io_pairs = (
      [({}, {})] if len(self.io_value_pairs) <= 0 else self.io_value_pairs
    )

    for io_pair in io_pairs:
      self._state['term'].reset_stream_lists()

      i, o = io_pair

      x_args = None if 'x' not in i else i['x']
      x_stdi = [] if 'stdi' not in i else i['stdi']
      self._state['term'].input_stream_as_list += x_stdi # Load into test terminal
      
      y_return = None if 'y' not in o else o['y']
      y_stdo = [] if 'stdo' not in o else o['stdo']

      z_return = self.run(x_args)
      z_stdo = self._state['term'].output_stream_as_list

      xyz = {
        'x_args': x_args,
        'x_stdi': x_stdi,
        'y_return': y_return,
        'y_stdo': y_stdo,
        'z_return': z_return,
        'z_stdo': z_stdo,
      }

      # Only do this comparison if the expected return is not none.
      if y_return is not None:
        # Test y_return == z_return
        if not self.equality_operator(y_return, z_return):
        # if y_return != z_return:
          return report.fail(
            self.name,
            xyz,
            'y_return != z_return',
            test_start_time,
            max_filename_length
          )

      # Test y_stdo == z_stdo
      if y_stdo != z_stdo:
        return report.fail(
          self.name,
          xyz,
          'y_stdo != z_stdo',
          test_start_time,
          max_filename_length
        )

      # Test z_return satisfies output properties
      for output_property in self._output_properties:
        if isinstance(output_property, Funk):
          if not output_property.run(z_return):
            return report.property_failure(
              self.name,
              f'{output_property.name}({z_return})',
              time_started=test_start_time,
              max_filename_length=max_filename_length
            )
        else:
          raise TypeError('property must be written as a Funk')

    if self.custom_tests is not None:
      for (custom_test_index, custom_test) in enumerate(self.custom_tests):
        if not custom_test():
          return report.fail(
            self.name,
            xyz,
            mode=f'custom_tests[{custom_test_index}]()',
            time_started=test_start_time,
            max_filename_length=max_filename_length
          )

    return report.success(self.name, test_start_time, max_filename_length)

  def test(self, max_filename_length):

    if self.enabled:
      self._state['term'].mode = 'test'
      # Start Timer
      _test = f"{'TEST':^8}"
      print(f'|{tc.warning(_test)}| {tc.info(self.name)} ', end='')
      result = self._test(max_filename_length)
      self._state['term'].mode = 'run'
      return result
    else:
      _skip = f"{'SKIP':^8}"
      print(f'|{tc.bright_magenta(_skip)}| {tc.bright_magenta(self.name)} ')
      return True

  __del__ = lambda self: self._tear_down(self._state)

  __call__ = lambda self, x=None: self._run(self._state, x)
